package org.jastadd.ragconnect.compiler;

import org.jastadd.ragconnect.ast.Document;
import org.jastadd.ragconnect.ast.ListElement;
import org.jastadd.ragconnect.ast.MappingElement;

/**
 * Testing Ros2Rag without parser.
 *
 * @author rschoene - Initial contribution
 */
public class SimpleMain {

  private static void printManualYAML() {
    Document doc = new Document();
    MappingElement root = new MappingElement();
    MappingElement firstLevel = new MappingElement();
    firstLevel.put("server", "tcp://localhost:1883");
    firstLevel.put("robot_speed_factor", ".7");

    MappingElement theTopics = new MappingElement();
    theTopics.put("robotConfig", "robotconfig");
    theTopics.put("trajectory", "trajectory");
    theTopics.put("nextStep", "ros2rag/nextStep");
    firstLevel.put("topics", theTopics);

    firstLevel.put("zone_size", "0.5");

    ListElement theZones = new ListElement();
    theZones.add("1 1");
    theZones.add("0 1");
    theZones.add("-1 1");
    firstLevel.put("zones", theZones);

    MappingElement pandaParts = new MappingElement();
    MappingElement thePanda = new MappingElement();
    thePanda.put("Link0", "panda_link0");
    thePanda.put("Link1", "panda_link1");
    thePanda.put("Link2", "panda_link2");
    thePanda.put("Link3", "panda_link3");
    thePanda.put("Link4", "panda_link4");
    thePanda.put("Link5", "panda_link5");
    thePanda.put("Link6", "panda_link6");
    thePanda.put("RightFinger", "panda_rightfinger");
    thePanda.put("LeftFinger", "panda_leftfinger");
    pandaParts.put("panda", thePanda);
    firstLevel.put("parts", pandaParts);

    MappingElement endEffectorParts = new MappingElement();
    MappingElement endEffector = new MappingElement();
    endEffector.put("EndEffector", "panda_hand");
    endEffectorParts.put("panda", endEffector);
    firstLevel.put("end_effectors", endEffectorParts);

    ListElement theGoalPoses = new ListElement();
    theGoalPoses.add(makePose("0.4 0.4 0.3"));
    theGoalPoses.add(makePose("-0.4 0.4 0.3"));
    theGoalPoses.add(makePose("-0.4 -0.4 0.3"));
    theGoalPoses.add(makePose("0.4 0.4 0.3"));
    theGoalPoses.add(makePose("-0.4 0.4 0.3"));
    theGoalPoses.add(makePose("0.4 0.4 0.3"));
    firstLevel.put("goal_poses", theGoalPoses);

    root.put("panda_mqtt_connector", firstLevel);
    doc.setRootElement(root);

    System.out.println(doc.prettyPrint());
  }

  private static MappingElement makePose(String position) {
    MappingElement goalPose = new MappingElement();
    goalPose.put("position", position);
    goalPose.put("orientation", "1 1 0 0");
    goalPose.put("work", "20000");
    return goalPose;
  }

  public static void main(String[] args) {
    printManualYAML();
  }
}
